//======================================================================
//  GivParser.cc - A parser for the giv file format.
//
//  Dov Grobgeld <dov.grobgeld@gmail.com>
//  Sun Nov  4 23:13:25 2007
//----------------------------------------------------------------------

#include <math.h>
#include "GivParser.h"
#include "giv_data.h"
#include "plis/plis.h"

using namespace plis;

enum
{
  STRING_NOP,
  STRING_DRAW,
  STRING_COMMENT,
  STRING_MOVE,
  STRING_QUIVER,
  STRING_TEXT,
  STRING_FONT,
  STRING_DASH,
  STRING_ARROW,
  STRING_CHANGE_COLOR,
  STRING_CHANGE_OUTLINE_COLOR,
  STRING_CHANGE_QUIVER_COLOR,
  STRING_CHANGE_QUIVER_SCALE,
  STRING_CHANGE_LINE_WIDTH,
  STRING_CHANGE_MARKS,
  STRING_CHANGE_NO_LINE,
  STRING_CHANGE_SCALE_MARKS,
  STRING_CHANGE_MARK_SIZE,
  STRING_CHANGE_TEXT_SIZE,
  STRING_CHANGE_LINE,
  STRING_CHANGE_NO_MARK,
  STRING_CHANGE_POLYGON,
  STRING_BALLOON,
  STRING_IMAGE_REFERENCE,
  STRING_MARKS_REFERENCE,
  STRING_LOW_CONTRAST,
  STRING_VFLIP,
  STRING_NO_VFLIP,
  STRING_HFLIP,
  STRING_NO_HFLIP,
  STRING_PATH_NAME,
  STRING_SET_NAME,
  STRING_SET_XUNIT_TEXT,
  STRING_SET_YUNIT_TEXT,
  STRING_SET_LARGE_PIXELS,
  STRING_SET_TITLE,
  STRING_STYLE,
  STRING_DEF_STYLE,
  STRING_HIDE,
  STRING_IGNORE
};

#define COLOR_NONE 0xfffe

#define NCASE(s) if (!g_ascii_strcasecmp(s, S_))
#define CASE(s) if (!strcmp(s, S_))

GdkColor set_colors[] =
  { {0, 0xffff, 0, 0},
    {0, 0, 0xffff, 0},  
    {0, 0, 0, 0xffff},
    {0, 0xffff, 0xffff, 0},
    {0, 0, 0xffff, 0xffff},
    {0, 0xffff, 0, 0xffff} };
gint nmarks_colors = 6;

GivParser::GivParser()
{
    cb_file_reference = NULL;
    clear();
    quiver_scale = 1.0;
}

int
GivParser::parse_file(const char *filename)
{
    int ret = 0;

    giv_dataset_t *marks = NULL;
    gboolean is_new_set;
    int num_sets = giv_datasets.size();
    int linenum = 0;
    
    FILE *GIV;

    GIV = fopen(filename, "r");
    if (!GIV)
	return -1;

    is_new_set = TRUE;
    while(1) {
	char S_[256];
	int len;
	
	linenum++;
	fgets(S_, sizeof(S_), GIV);
	len = strlen(S_);

        if (feof(GIV))
            break;
#if 0
	if (len<=1)
	    continue;
#endif
	
	// Get rid of CR and LF at end of line
	while (len>0 && (S_[len-1] == '\r' || S_[len-1] == '\n')) {
	    S_[len-1] = 0;
	    len--;
	}
	
	if (is_new_set || marks==NULL) {
	    marks = new_giv_dataset(num_sets);
	    marks->color = set_colors[num_sets % nmarks_colors];
	    marks->file_name = g_strdup(filename);
            giv_datasets.push_back(marks);
	    
	    is_new_set = FALSE;
	    num_sets++;
	}
	
	if (len == 0) {
	    if (marks && ((GArray*)marks->points)->len > 0)
		is_new_set++;
	    continue;
	}

        giv_marks_data_add_line(marks, S_, filename, linenum);
    }
    fclose(GIV);

    /* Get rid of empty data sets */
    if (marks && marks->points->len == 0) {
	giv_datasets.pop_back();
	free_giv_data_set(marks);
	marks = NULL;
    }
    
    return ret;
}


/*======================================================================
//
// General string handling in C. These functions should be replaced
// by something in glib.
//
//----------------------------------------------------------------------*/
int
string_count_words (const char *string)
{
  int nwords = 0;
  const char *p = string;
  int in_word = 0;
  while (*p)
    {
      if (!in_word)
	{
	  if ((*p != ' ') && (*p != '\n') && (*p != '\t'))
	    in_word = 1;
	}
      else if (in_word)
	{
	  if ((*p == ' ') || (*p == '\n') || (*p != '\t'))
	    {
	      in_word = 0;
	      nwords++;
	    }
	}
      p++;
    }
  if (in_word)
    nwords++;
  return nwords;
}

char *
string_strdup_word (const char *string, int idx)
{
  const char *p = string;
  int in_word = 0;
  int word_count = -1;
  int nchr = 1;
  char *word = NULL;
  const char *word_start = string;

  /* printf("p = 0x%x\n", p); */
  while (*p)
    {
      /* printf("%c\n", *p); fflush(stdout); */
      if (!in_word)
	{
	  if (*p != ' ' && *p != '\n' && *p != '\t')
	    {
	      in_word = 1;
	      word_count++;
	      nchr = 1;
	      word_start = p;
	    }
	}
      else if (in_word)
	{
	  if (*p == ' ' || *p == '\n' || *p == '\t')
	    {
	      if (idx == word_count)
		break;
	      in_word = 0;
	    }
	}
      nchr++;
      p++;
    }
  if (in_word)
    {
      word = g_new (char, nchr);
      strncpy (word, word_start, nchr - 1);
      word[nchr - 1] = 0;
    }
  return word;
}

char *
string_strdup_rest (const char *string, int idx)
{
  const char *p = string;
  int in_word = 0;
  int word_count = -1;
  int nchr = 1;
  char *word = NULL;
  const char *word_start = string;

  /* printf("p = 0x%x\n", p); */
  while (*p)
    {
      /* printf("%c\n", *p); fflush(stdout); */
      if (!in_word)
	{
	  if (*p != ' ' && *p != '\n' && *p != '\t')
	    {
	      in_word = 1;
	      word_count++;
	      nchr = 1;
	      word_start = p;
	    }
	}
      else if (in_word)
	{
	  if (*p == ' ' || *p == '\n' || *p == '\t')
	    {
	      if (idx == word_count)
		break;
	      in_word = 0;
	    }
	}
      nchr++;
      p++;
    }

  if (in_word)
    {
      nchr = strlen (word_start);
      word = g_new (char, nchr + 1);
      strncpy (word, word_start, nchr);
      word[nchr] = 0;
    }
  return word;
}

int
string_to_atoi (const char *string, int idx)
{
  char *word = string_strdup_word (string, idx);
  int value = atoi (word);
  g_free (word);

  return value;
}

gdouble
string_to_atof (const char *string, int idx)
{
  char *word = string_strdup_word (string, idx);
  gdouble value = atof (word);
  g_free (word);

  return value;
}


/*======================================================================
//  Classify a string.
//----------------------------------------------------------------------
*/
gint
giv_parse_string (const char *string, const char *fn, gint linenum)
{
  gint type = -1;
  gchar first_char = string[0];
  gchar *first_word;

  /* Shortcut for speeding up drawing */
  if (first_char >= '0' && first_char <= '9')
    return STRING_DRAW;

  first_word = string_strdup_word (string, 0);

  if (first_char == '#')
    type = STRING_COMMENT;
  else if (first_char == '$')
    {
      char *S_ = first_word;
      NCASE ("$lw")
      {
	type = STRING_CHANGE_LINE_WIDTH;
      }
      NCASE("$balloon")
      {
          type = STRING_BALLOON;
      }
      NCASE ("$color")
      {
	type = STRING_CHANGE_COLOR;
      }
      NCASE ("$outline_color")
      {
	type = STRING_CHANGE_OUTLINE_COLOR;
      }
      NCASE ("$quiver_color")
      {
	type = STRING_CHANGE_QUIVER_COLOR;
      }
      NCASE ("$quiver_scale")
      {
	type = STRING_CHANGE_QUIVER_SCALE;
      }
      NCASE ("$marks")
      {
	type = STRING_CHANGE_MARKS;
      }
      NCASE ("$noline")
      {
	type = STRING_CHANGE_NO_LINE;
      }
      NCASE ("$scale_marks")
      {
	type = STRING_CHANGE_SCALE_MARKS;
      }
      NCASE ("$path")
      {
	type = STRING_PATH_NAME;
      }
      NCASE ("$mark_size")
      {
	type = STRING_CHANGE_MARK_SIZE;
      }
      NCASE ("$text_size")
      {
	type = STRING_CHANGE_TEXT_SIZE;
      }
      NCASE ("$font")
      {
          type = STRING_FONT;
      }
      NCASE ("$nomark")
      {
	type = STRING_CHANGE_NO_MARK;
      }
      NCASE ("$line")
      {
	type = STRING_CHANGE_LINE;
      }
      NCASE ("$image")
      {
	type = STRING_IMAGE_REFERENCE;
      }
      NCASE ("$polygon")
      {
	type = STRING_CHANGE_POLYGON;
      }
      NCASE ("$marks_file")
      {
	type = STRING_MARKS_REFERENCE;
      }
      NCASE ("$low_contrast")
      {
	type = STRING_LOW_CONTRAST;
      }
      NCASE("$vflip")
        {
          type = STRING_VFLIP;
        }
      NCASE("$novflip")
        {
          type = STRING_NO_VFLIP;
        }
      NCASE("$hflip")
        {
          type = STRING_HFLIP;
        }
      NCASE("$nohflip")
        {
          type = STRING_NO_HFLIP;
        }
      NCASE("$linedash")
        {
          type = STRING_DASH;
        }
      NCASE("$arrow") {
          type = STRING_ARROW;
      }
      NCASE("$def_style")
	{
	  type = STRING_DEF_STYLE;
	}
      NCASE("$style")
	{
	  type = STRING_STYLE;
	}
      NCASE ("$title")
      {
	type = STRING_SET_TITLE;
      }
      NCASE ("$name")
      {
	type = STRING_SET_NAME;
      }
      NCASE ("$hide")
      {
	type = STRING_HIDE;
      }
#if 0
      if (type == -1)
	{
	  fprintf (stderr, "Unknown parameter %s in file %s line %d!\n", S_,
		   fn, linenum);
	}
#endif
    }
  else if (((first_char >= 'a' && first_char <= 'z')
	    || (first_char >= 'A' && first_char <= 'Z'))
	   && first_word[strlen (first_word) - 1] == ':')
    {
      char *S_ = first_word;

      while(1) /* So that we may break */
	{
	  NCASE("Title:") { type = STRING_SET_TITLE; break;}
	  NCASE("TitleText:") { type = STRING_SET_TITLE; break;}
	  NCASE("XUnitText:") { type = STRING_SET_XUNIT_TEXT; break; }
	  NCASE("YUnitText:") { type = STRING_SET_YUNIT_TEXT; break; }
	  NCASE("LargePixels:") { type = STRING_SET_LARGE_PIXELS; break; }

	  printf("Unsupported keyword=%s\n", S_);
	  // TBD - Recognize more xgraph keywords...
	  type = STRING_NOP;

	  break;
	}

    }
  else if (first_char == 'M' || first_char == 'm')
    {
      type = STRING_MOVE;
    }
  else if (first_char == 'Q' || first_char == 'q')
      type = STRING_QUIVER;
  else if (first_char == 'T' || first_char == 't')
    {
      type = STRING_TEXT;
    }
  else if (first_char == '"')
    {
      type = STRING_SET_NAME;
    }
  else
    {
      type = STRING_DRAW;
    }

  g_free (first_word);

  return type;
}

gint
giv_parse_mark_type (const char *S_, const gchar * fn, gint linenum)
{
  NCASE ("circle") return MARK_TYPE_CIRCLE;
  NCASE ("fcircle") return MARK_TYPE_FCIRCLE;
  NCASE ("square") return MARK_TYPE_SQUARE;
  NCASE ("fsquare") return MARK_TYPE_FSQUARE;
  NCASE ("pixel") return MARK_TYPE_PIXEL;
  fprintf (stderr, "Unknown mark %s in file %s line %d\n", S_, fn, linenum);
  return MARK_TYPE_CIRCLE;
}


int
GivParser::giv_marks_data_add_line(giv_dataset_t *marks,
                                   const char *line,
                                   const char *filename,
                                   int linenum)
{
    int type;
    const char *S_ = line; 
    char dummy[256];
    point_t p;
    // Only take marks size into account if we are scaling the marks!
    double ms2 = 0.5*marks->mark_size * marks->do_scale_marks; 

    /* Parse the line */
    type = giv_parse_string(line, filename, linenum);
    switch (type) {
    case STRING_COMMENT:
        break;
    case STRING_DRAW:
    case STRING_MOVE:
    case STRING_QUIVER:
        if (type == STRING_DRAW) {
            sscanf(S_, "%lf %lf", &p.data.point.x, &p.data.point.y);
            if (marks->points->len == 0)
                p.op = OP_MOVE;
            else
                p.op = OP_DRAW;
        }
        else {
            sscanf(S_, "%s %lf %lf", dummy, &p.data.point.x, &p.data.point.y);
            if (type == STRING_QUIVER) {
                p.op = OP_QUIVER;
                marks->has_quiver = TRUE;
            }
            else
                p.op = OP_MOVE;
        }

        /* Find marks bounding box */
        if (p.data.point.x < global_mark_min_x)
            global_mark_min_x = p.data.point.x - ms2;
        else if (p.data.point.x > global_mark_max_x)
            global_mark_max_x = p.data.point.x + ms2;
        if (p.data.point.y < global_mark_min_y)
            global_mark_min_y = p.data.point.y - ms2;
        else if (p.data.point.y > global_mark_max_y)
            global_mark_max_y = p.data.point.y + ms2;
	    
        g_array_append_val(marks->points, p);
        break;
    case STRING_TEXT:
        {
            text_mark_t *tm = (text_mark_t*)g_new(text_mark_t, 1);
            sscanf(S_, "%s %lf %lf", dummy, &tm->x, &tm->y);
            tm->string = string_strdup_rest(S_, 3);
            p.op = OP_TEXT;
            p.data.point.x = tm->x;
            p.data.point.y = tm->y;
            p.data.text_object = tm;
            g_array_append_val(marks->points, p);
        }
        break;
    case STRING_CHANGE_LINE_WIDTH:
        marks->line_width = string_to_atof(S_, 1);
        break;
    case STRING_BALLOON:
        {
            char *s = string_strdup_rest(S_,1);
            if (marks->balloon_string.length())
                marks->balloon_string+= "\n";
            marks->balloon_string += (slip)s;
            g_free(s);
            
            break;
        }
    case STRING_IMAGE_REFERENCE:
        {
            char *image_filename = string_strdup_word(S_, 1);

            if (cb_file_reference)
                (*cb_file_reference)(image_filename, cb_file_reference_data);
            // In contrast to other references image commands are carried out immediately!
            // salfw_viewer_load_file(image_filename, 0);
                  
            g_free(image_filename);

            break;
        }
    case STRING_VFLIP:
        if (cb_set_orientation)
            (*cb_set_orientation)(ORIENTATION_UNDEF, ORIENTATION_FLIP, cb_set_orientation_data);
        break;
    case STRING_NO_VFLIP:
        if (cb_set_orientation)
            (*cb_set_orientation)(ORIENTATION_UNDEF, ORIENTATION_UNFLIP, cb_set_orientation_data);
        break;
    case STRING_HFLIP:
        if (cb_set_orientation)
            (*cb_set_orientation)(ORIENTATION_FLIP, ORIENTATION_UNDEF, cb_set_orientation_data);
        break;
    case STRING_NO_HFLIP:
        if (cb_set_orientation)
            (*cb_set_orientation)(ORIENTATION_UNFLIP, ORIENTATION_UNDEF, cb_set_orientation_data);
        break;
    case STRING_MARKS_REFERENCE:
        {
#if 0
            char *marks_filename = string_strdup_word(S_, 1);
                
            // Todo: Make image relative to the marks list
            g_ptr_array_add(mark_file_name_list, marks_filename);
                
#endif
            break;
        }
    case STRING_LOW_CONTRAST:
        {
#if 0
            giv_current_transfer_function = TRANS_FUNC_LOW_CONTRAST;
#endif
            break;
        }
    case STRING_DASH:
        {
            // Get string and split it on commas
            char *p;
            char *dash_string = string_strdup_word(S_,1);
            int i, num_dashes;
            
            // Replace comma with spaces for easier parsing
            p = dash_string;
            while((p=g_strstr_len(p, strlen(dash_string), ","))) 
                *p = ' ';
            
            num_dashes = string_count_words(dash_string);
            marks->num_dashes = num_dashes;
            marks->dashes = g_new0(double, num_dashes);
            for (i=0; i<num_dashes; i++)
                marks->dashes[i] = string_to_atof(dash_string, i);
            free(dash_string);
            
            break;
        }
    case STRING_ARROW:
        {
            char *arrow_string = string_strdup_word(S_,1);
            if (strcmp(arrow_string, "start")==0)
                marks->arrow_type = ARROW_TYPE_START;
            else if (strcmp(arrow_string, "both")==0)
                marks->arrow_type = ARROW_TYPE_BOTH;
            else 
                marks->arrow_type = ARROW_TYPE_END;
            break;
        }
    case STRING_CHANGE_NO_LINE:
        marks->do_draw_lines = FALSE;
        break;
    case STRING_CHANGE_POLYGON:
        marks->do_draw_polygon = TRUE;
        break;
    case STRING_CHANGE_LINE:
        marks->do_draw_lines = TRUE;
        break;
    case STRING_CHANGE_NO_MARK:
        marks->do_draw_marks = FALSE;
        break;
    case STRING_CHANGE_MARK_SIZE:
        marks->mark_size = string_to_atof(S_, 1);
        break;
    case STRING_CHANGE_TEXT_SIZE:
        marks->text_size = string_to_atof(S_, 1);
        break;
    case STRING_FONT:
        if (marks->font_name)
            g_free(marks->font_name);
        marks->font_name = string_strdup_rest(S_, 1);
        break;
    case STRING_CHANGE_COLOR:
        {
            char *color_name = string_strdup_word(S_, 1);
            GdkColor color = {0,0,0,0};
            if ((slip)color_name == "none")
                marks->color.pixel = COLOR_NONE;
            else {
                if (gdk_color_parse(color_name,&color))
                    marks->color = color;
            }
            g_free(color_name);
            break;
        }
    case STRING_CHANGE_OUTLINE_COLOR:
        {
            char *color_name = string_strdup_word(S_, 1);
            GdkColor color = {0,0,0,0};
            if ((slip)color_name == "none")
                marks->outline_color.pixel = COLOR_NONE;
            else {
                if (gdk_color_parse(color_name,&color))
                    marks->outline_color = color;
            }
            marks->do_draw_polygon_outline = TRUE;
            g_free(color_name);
            break;
        }
    case STRING_CHANGE_QUIVER_COLOR:
        {
            char *color_name = string_strdup_word(S_, 1);
            GdkColor color = {0,0,0,0};
            if ((slip)color_name == "none")
                marks->quiver_color.pixel = COLOR_NONE;
            else {
                if (gdk_color_parse(color_name,&color))
                    marks->quiver_color = color;
            }
            g_free(color_name);
            break;
        }
    case STRING_CHANGE_QUIVER_SCALE:
        marks->quiver_scale = string_to_atof(S_, 1);

        // remember last value
        quiver_scale = marks->quiver_scale;
        break;
    case STRING_CHANGE_MARKS:
        {
            char *mark_name = string_strdup_word(S_, 1);
            marks->do_draw_marks = TRUE;
	      
            marks->mark_type = giv_parse_mark_type(mark_name, filename, linenum);
		
            g_free(mark_name);
            break;
        }
    case STRING_CHANGE_SCALE_MARKS:
        if (string_count_words(S_) == 1)
            marks->do_scale_marks = 1;
        else
            marks->do_scale_marks = string_to_atoi(S_, 1);
        break;
    case STRING_PATH_NAME:
        if (marks->path_name)
            g_free(marks->path_name);
        marks->path_name = string_strdup_rest(S_, 1);
        break;
    case STRING_DEF_STYLE:
        {
            char *style_name = string_strdup_word(S_, 1);
            char *style_string = string_strdup_rest(S_, 2);
            giv_style_add_string(style_name, style_string);
            g_free(style_name);
            g_free(style_string);
        }
        break;
    case STRING_STYLE:
        {
            char *style_name = string_strdup_word(S_, 1);
            giv_set_props_from_style(marks, style_name);
            g_free(style_name);
            break;
        }
    case STRING_HIDE:
        {
            marks->is_visible = false;
            break;
        }
    default:
        ;
#if 0
    default:
        printf("Unknown string: %s\n", S_);
#endif
    }
    return 0;
}

void
GivParser::giv_style_add_string(const char* style_name,
                                const char* style_string)
{
    llip matches;

    if (slip(style_string).m("^(\\S+)\\s+(.*)", matches)) {
        // Search if the same property is already defined, then overwrite it.
        slip key = matches[1];
        for (int i=0; i< (int)style_hash[style_name].size(); i++) {
            if (style_hash[style_name][i].m(key, "i")) {
                style_hash[style_name][i] = style_string;
                return;
            }
        }
        
        style_hash[style_name].push_back(style_string);
    }
}

void
GivParser::giv_set_props_from_style(giv_dataset_t *marks,
                                    const char *style_name)
{
    if (style_hash.find(style_name) == style_hash.end()) {
        return;
    }

    /* Get arary of properties */
    llip& style_array = style_hash[style_name];

    int prop_idx;

    /* Loop over all style properties and parse them */
    for (prop_idx=0; prop_idx < style_array; prop_idx++) {
        slip style_string = style_array[prop_idx];
        llip args = style_string.split();

        slip S_ = args.shift();
        S_.tr("A-Z","a-z"); // lower case

        CASE("lw") {
            marks->line_width = args.shift().atof();
            continue;
        }
        CASE("color") {
            slip color_name = args.shift();
            GdkColor color = {0,0,0,0};
            if (color_name == "none")
                marks->color.pixel = COLOR_NONE;
            else {
                if (gdk_color_parse(color_name.c_str(),&color))
                    marks->color = color;
            }
            continue;
        }
        CASE("outline_color") {
            slip color_name = args.shift();
            GdkColor color = {0,0,0,0};
            if (gdk_color_parse(color_name,&color))
                marks->outline_color = color;
            marks->do_draw_polygon_outline = TRUE;
            continue;
        }
        CASE("marks") {
            slip mark_name = args.shift();
            marks->do_draw_marks = TRUE;
	  
            marks->mark_type = giv_parse_mark_type(mark_name, style_name, prop_idx);
	  
            continue;
        }
        CASE("noline") {
            marks->do_draw_lines = FALSE;
            continue;
        }
        CASE("line") {
            marks->do_draw_lines = TRUE;
            continue;
        }
        CASE("scale_marks") {
            if (args.length() > 0)
                marks->do_scale_marks = args[1].atoi();
            else 
                marks->do_scale_marks = 1;
            continue;
        }
        CASE("mark_size") {
            if (args.length() > 0)
                marks->mark_size = args[1].atof();
            continue;
        }
        CASE("nomark") {
            marks->do_draw_marks = FALSE;
            continue;
        }
        CASE("polygon") {
            marks->do_draw_polygon = TRUE;
            continue;
        }
        CASE("arrow") {
            if (args.length() < 1
                || args[1] == "end") {
                marks->arrow_type = ARROW_TYPE_END;
            }
            else if (args[1] == "start") 
                marks->arrow_type = ARROW_TYPE_START;
            else if (args[1] == "both")
                marks->arrow_type = ARROW_TYPE_BOTH;
        }
        printf("Unknown option %s!\n", (const char*)S_);
    }
}

void
GivParser::set_reference_callback(file_reference_t fr, gpointer user_data)
{
    this->cb_file_reference = fr;
    this->cb_file_reference_data = user_data;
}

void
GivParser::set_orientation_callback(set_orientation_t cb, gpointer user_data)
{
    this->cb_set_orientation = cb;
    this->cb_set_orientation_data = user_data;
}

void
GivParser::clear()
{
    for (int i=0; i<(int)giv_datasets.size(); i++)
        free_giv_data_set(giv_datasets[i]);
        
    giv_datasets.clear();
    global_mark_max_x = global_mark_max_y = -1e30; // -HUGE
    global_mark_min_x = global_mark_min_y = 1e30; // HUGE
}

void GivParser::get_data_bbox(double &min_x, double& min_y,
                              double &max_x, double& max_y)
{
    min_x = global_mark_min_x;
    min_y = global_mark_min_y;
    max_x = global_mark_max_x;
    max_y = global_mark_max_y;
}

// Propagate quiver scale
void GivParser::set_quiver_scale(double quiver_scale)
{
    for (int i=0; i<(int)giv_datasets.size(); i++)
        giv_datasets[i]->quiver_scale = quiver_scale;
    
}
